@@ -29,6 +29,9 @@
 # undef  USE_PRELOAD
 # define CHECK_FORDEVPTS
 #endif
+#ifdef __FreeBSD__
+#undef CHECK_FORDEVPTS
+#endif
 #include <stdio.h>
 #include <termios.h>
 #include <signal.h>
@@ -39,7 +42,9 @@
 #include <sys/ioctl.h>
 #include <sys/socket.h>
 #include <sys/un.h>
+#ifndef __FreeBSD__
 #include <sys/sysinfo.h>
+#endif
 #include <sys/stat.h>
 #include <sys/mman.h>
 #ifdef CHECK_FORDEVPTS
@@ -78,6 +83,15 @@
 # define attribute(attr)	__attribute__(attr)
 #endif
 
+#ifdef __FreeBSD__
+# define TEMP_FAILURE_RETRY(expression) \
+  (__extension__                                                              \
+    ({ long int __result;                                                     \
+       do __result = (long int) (expression);                                 \
+       while (__result == -1L && errno == EINTR);                             \
+       __result; }))
+#endif
+
 typedef enum _boolean {false, true} boolean;
 extern char *optarg;
 extern int optind;
@@ -103,6 +117,19 @@ static sig_atomic_t signaled;
 const char *initddir = "/etc/init.d";
 const char *etcdir = "/etc";
 
+#define LEGACY_DEPENDENCY_PATH "/etc/init.d/."
+#define DEPENDENCY_PATH "/var/lib/insserv/"
+char *dependency_path = DEPENDENCY_PATH;
+#ifndef PATH_MAX
+#define PATH_MAX 2048
+#endif
+#ifndef TRUE
+#define TRUE 1
+#endif
+#ifndef FALSE
+#define FALSE 0
+#endif
+
 #define PBUF_SIZE	8192
 #define PRUNNING	0x0001
 #define PFINISHED	0x0002
@@ -288,8 +315,13 @@ void writebuf(struct prg *p)
 static int checksystem(const int par, const boolean start, const boolean limit)
 {
   const      int pg_size       = sysconf(_SC_PAGESIZE);
+  #ifdef __FreeBSD__
+  const long int minphys_bytes = (2<<21);
+  const long int avphys_pg = 10;
+  #else
   const long int minphys_bytes = (sysconf(_SC_LONG_BIT) > 32L) ? (2<<22) : (2<<21);
   const long int avphys_pg     = sysconf(_SC_AVPHYS_PAGES);
+  #endif
   long int minphys_pg;
   unsigned long int prcs_run, prcs_blked;
   int newpar;
@@ -419,7 +451,9 @@ void run(struct prg *p)
 
   p->len = 0;
   p->pid = (pid_t)0;
+  #ifndef __FreeBSD__
   p->fd = getpt();
+  #endif
   if (p->fd <= 0)
     {
       p->fd = 0;
@@ -551,8 +585,8 @@ void run(struct prg *p)
 
   if (run_mode)
     {
-      char path[128];
-      snprintf(path, sizeof(path), "%s/%s", initddir, p->name);
+      char path[PATH_MAX];
+      snprintf(path, PATH_MAX, "%s/%s", initddir, p->name);
       execlp(path, p->arg0, arg, (char *)0);
     }
   else if (arg)
@@ -595,8 +629,8 @@ int run_single(const char *prg, const char *arg0, int spl)
       closeall();
       if (run_mode)
 	{
-	  char path[128];
-	  snprintf(path, sizeof(path), "%s/%s", initddir, prg);
+	  char path[PATH_MAX];
+	  snprintf(path, PATH_MAX, "%s/%s", initddir, prg);
 	  execlp(path, arg0 ? arg0 : path, arg, (char *)0);
 	}
       else if (arg)
@@ -703,7 +737,7 @@ void detach(struct prg *p, const int store)
 
   if (-1 == fcntl(p->fd, F_SETFL, flags))
     {
-      perror("unable to set FNONBLOCK flag on program file descriptior");
+      perror("unable to set FNONBLOCK flag on program file descriptor");
     }
   while ((r = read(p->fd, p->buf, sizeof(p->buf))))
     {
@@ -722,7 +756,7 @@ void detach(struct prg *p, const int store)
   flags &= ~FNONBLOCK;
   if (-1 == fcntl(p->fd, F_SETFL, flags))
     {
-      perror("unable to unset FNONBLOCK flag on program file descriptior");
+      perror("unable to unset FNONBLOCK flag on program file descriptor");
     }
   if (r == -1 && errno == EWOULDBLOCK)
     {
@@ -732,6 +766,7 @@ void detach(struct prg *p, const int store)
 	  (void)sigfillset(&nmask);
 	  sigprocmask(SIG_UNBLOCK, &nmask, NULL);
 
+	  (void)signal(SIGINT,  SIG_DFL);   /* some interactive services may be interruptable */
 	  (void)signal(SIGHUP,  SIG_DFL);
 	  (void)signal(SIGQUIT, SIG_DFL);
 	  (void)signal(SIGSEGV, SIG_DFL);
@@ -826,9 +861,13 @@ void usage(int status)
   fprintf(stderr, "       startpar -v\n");
   fprintf(stderr, "           show version number\n");
   fprintf(stderr, "general options:\n");
+  fprintf(stderr, "       -l use legacy /etc/init.d path for Makefile-style scripts\n");
+  fprintf(stderr, "          The default is to use the location /var/lib/insserv\n");
   fprintf(stderr, "       -p parallel tasks\n");
   fprintf(stderr, "       -t I/O timeout\n");
   fprintf(stderr, "       -T global I/O timeout\n");
+  fprintf(stderr, "       -f copy stdin to stdout without running any programs\n");
+  fprintf(stderr, "       -e set the location of etc directory (default is /etc)\n");
   exit(status);
 }
 
@@ -849,6 +888,8 @@ int main(int argc, char **argv)
   char *run_level = getenv("RUNLEVEL");
   char *splashopt = 0;
   sigset_t nmask, omask, smask;
+  int status;     /* check if parsing makefile worked */
+  int using_legacy_path = FALSE;      /* use new makefile path by default */
 
   detect_consoles();
 
@@ -861,7 +902,7 @@ int main(int argc, char **argv)
   numcpu = sysconf(_SC_NPROCESSORS_ONLN);
   myname = argv[0];
 
-  while ((c = getopt(argc, argv, "fhp:t:T:a:M:P:R:S:vi:e:d:")) != EOF)
+  while ((c = getopt(argc, argv, "flhp:t:T:a:M:P:R:S:vi:e:d:")) != EOF)
     {
       switch(c)
         {
@@ -880,12 +921,16 @@ int main(int argc, char **argv)
 	case 'a':
 	  arg = optarg;
 	  break;
-	case 'e':
-	  etcdir = optarg;
-	  break;
 	case 'd':
 	  initddir = optarg;
 	  break;
+        case 'e':
+          etcdir = optarg;
+          break;
+        case 'l':
+          dependency_path = LEGACY_DEPENDENCY_PATH;
+          using_legacy_path = TRUE;
+          break;
 	case 'M':
 	  run_mode = optarg;
 	  break;
@@ -938,7 +983,7 @@ int main(int argc, char **argv)
     }
   if (run_mode)
     {
-      char makefile[64];
+      char makefile[PATH_MAX];
       if (!strcmp(run_mode, "boot"))
 	arg = "start";
       else if (!strcmp(run_mode, "halt"))
@@ -957,8 +1002,22 @@ int main(int argc, char **argv)
 	  fprintf(stderr, "invalid run mode %s\n", run_mode);
 	  exit(1);
 	}
-      snprintf(makefile, sizeof(makefile), "%s/.depend.%s", initddir, run_mode);
-      parse_makefile(makefile);
+      snprintf(makefile, PATH_MAX, "%sdepend.%s", dependency_path, run_mode);
+      status = parse_makefile(makefile);
+      /* If the parse fails it is probably because the file does not
+         exist. Try alternative location. -- Jesse */
+      if ( (! status) && (! using_legacy_path) )
+      {
+         snprintf(makefile, PATH_MAX, "%sdepend.%s", LEGACY_DEPENDENCY_PATH, run_mode);
+         fprintf(stderr, "Trying to find makefile-style script at legacy location: %s\n",
+                         LEGACY_DEPENDENCY_PATH);
+         status = parse_makefile(makefile);
+         if (! status)
+            exit(1);
+      }
+      else if ( (! status) && (using_legacy_path) )
+         exit(1);      /* no other place to try so drop out */
+
       check_run_files(run_mode, prev_level, run_level);
 
       argc = tree_entries;			/* number of handled scripts */
@@ -1251,7 +1310,7 @@ int main(int argc, char **argv)
 #if DEBUG
       fprintf(stderr, "active = %d, maxfd = %d, last = %d\n", active, maxfd, last);
 #endif
-      if (active == 1 && last >= 0)
+      if (active == 1 && last >= 0 && num >= argc)
 	{
 	  p = prgs + last;
 	  if ((pid = waitpid(p->pid, &r, (maxfd < 0 ? 0 : WNOHANG)|WUNTRACED)) == 0)
